unit AsmUnit;

interface

uses Classes, SysUtils, TypeDefs;

const
  AsmPackedArraySize = 315;

{The array which stores the information for decoding a source line is in a
 packed format; unnecessary as far as Delphi is concerned, but like this in
 order to give me experience of a minimal array for future implementation
 within the memory available on an HC11 development board - specifically as
 part of a traditional FORTH compiler resident on the target system.

 The packed format is:

 Byte 1     Byte 2     Byte 3     Byte 4     Byte 5
 11111222   22333334   44445555   66666666   77--8888

 Key:
 1 : 1st char of mnemonic (5 bits)
 2 : 2nd char of mnemonic (5 bits)
 3 : 3rd char of mnemonic (5 bits)
 4 : 4th char of mnemonic (5 bits)
 All the mnemonic characters are derived by 'anding' the char with $1F.
 Two instructions in the HC11 set have 5 characters, but fortunately these
 two, BRCLR and BRSET, both have unique 4th characters so the fifth char
 can be ignored.
 5 : code for addressing mode (4 bits):
     tAddrMode = (amNone,amInherent,amImmediate,amDirect,amExtended,
                   amIndirectX,amIndirectY,amRelative,amDirective);
     Code generated by ord(AddrMode);
 6 : Main opcode byte (8 bits)
 7 : (2 bits) If zero, then no pre-byte. Otherwise pre-byte by a lookup tabe:
      case (b5 shr 6) of
        0 : OpcodeStr := OpStr;
        1 : OpcodeStr := '18' + OpStr;
        2 : OpcodeStr := '1A' + OpStr;
        3 : OpcodeStr := 'CD' + OpStr;
      end;
 8 : code for a format string describing how the operand(s) should be
     constructed (4 bits). This code is entered into a lookup table, see
     the definition of 'OperandFormats' below.
 }
type
  tAsmPackedRec   = array[1..5] of byte;
  tAsmPackedArray = array[1..AsmPackedArraySize] of tAsmPackedRec;
var
  AsmPA : tAsmPackedArray;
const
  OperandFormats : array[0..15] of Str4 =
   ('    ',    // Inherent instruction
    'D   ',    // Direct addressing
    'DM  ',    // Direct addressing, mask byte
    'DMR ',    // Direct addressing, mask byte, relative offset
    'FX  ',    // Forward offset
    'FXM ',    // Forward offset, mask byte
    'FXMR',    // Forward offset, mask byte, relative offset
    'HL  ',    // Extended addressing
    'I   ',    // Immediate addressing, one byte
    'JK  ',    // Immediate addressing, two bytes
    'R   ',    // Relative addressing
    'B   ',    // Byte string (FCB directive)
    'E   ',    // word        (EQU directive)
    'W   ',    // Word string (FDB directive)
    'C   ',    // Char string (FCC directive)
    'S   '     // Reserve space in memory
    );

const
  NumDirectives = 5;
  Directives : array[1..NumDirectives] of string =
    ('EQU,"","",DRV,,,ee,,,0,0',
     'FCB,"","",DRV,,,bytes,,,0,0',
     'FDB,"","",DRV,,,words,,,0,0',
     'FCC,"","",DRV,,,chars,,,0,0',
     'RMB,"","",DRV,,,space,,,0,0');

procedure AddUserLabel(S : string; W : word);
function  DeleteUserLabel(S : string) : boolean;
procedure MakeAsmPA(i : integer; S : string);
procedure MakeAsmBytes(    S : string; AddrMode : tAddrMode;
                       var AsmBytes : tAsmPackedRec);
function  GetAsmBytes(var AsmBytes : tAsmPackedRec;
                      var ArrayIndex : integer) : boolean;
procedure ExtractAsmInfo(    ArrayIndex : integer;
                         var AddrMode   : tAddrMode;
//                         var NumModes   : integer;
                         var OpcodeStr,FormatStr : string);
function  Asm1Check(const CList : tStringList;
                    var Command : tCommand) : tCmdLineComment;
function  Asm2Check(const CList     : tStringList;
                          CmdText     : string;
                          LastCommand : tCommand;
                   var   Command     : tCommand) : tCmdLineComment;
procedure AsmLocal1(var Command : tCommand);
procedure AsmLocal2(var Command : tCommand);
procedure Asm1(var Command : tCommand);
procedure Asm2(var Command : tCommand);
function  ExtConversion(    S : string;
                        var W : word) : tTokenResult;
procedure GetOperandPart(    OperandText : string;
                             PartNum     : integer;
                         var PartStr     : string);

implementation

uses Main, CmdUnit, LookupUnit, TalkerComUnit, SelectUnit;

type
  tAsmErr = (aeNone,
             aeIllegalCharInLabel,
             aeDuplicateLabel,
             aeLabelNotDefined,
             aeIllegalLabel,
             aeLabelMissing,
             aeUnrecogMnemonic,
             aeUnterminatedQuote,
             aeOperandOutOfRange,
             aeOperandError,
             aeBranchOutOfRange,
             aeNoMnemonic,
             aeNoOperand,
             aeIntErr1,
             aeIntErr2,
             aeIntErr3,
             aeIntErr4,
             aeIntErr5,
             aeIntErr6);

const
  AsmErrMsgs : array[tAsmErr] of string =
   ('',
    'Illegal character in label',
    'Duplicate label',
    'Label not defined',
    'Illegal label',
    'Label missing',
    'Unrecognized mnemonic',
    'Unterminated quotation',
    'Operand out of range',
    'Operand error',
    'Branch out of range',
    'Instruction mnemonic missing',
    'Operand(s) missing',
    'Internal error #1',
    'Internal error #2',
    'Internal error #3',
    'Internal error #4',
    'Internal error #5',
    'Internal error #6');

var
  AsmErr : tAsmErr;
  BPROTChecked : boolean;

function LabelCheck(S : string) : boolean;
var j : integer;
begin
  j := 1;
  while (j <= length(S)) and
        (upcase(S[j]) in ['_','0'..'9','A'..'Z']) do inc(j);
  Result := j > length(S);
end;

procedure AddUserLabel(S : string; W : word);
begin
  setlength(UserTable,length(UserTable)+1);
  with UserTable[high(UserTable)] do
  begin
    Name := S;
    Val  := W;
  end;
  if (SelectListing = slSymbols) and SelectForm.Visible
  then begin
    SelectForm.ListBox1.Items.Add(IntToHex(W,4) + '  ' + S);
    SelectForm.ListBox1.ItemIndex := SelectForm.ListBox1.Items.Count - 1;
  end;
end; // procedure AddUserLabel

function DeleteUserLabel(S : string) : boolean;
var
  i,j : integer;
begin
  Result := false;
  i := 0;
  while (i < length(UserTable)) and
        (uppercase(UserTable[i].Name) <> uppercase(S)) do inc(i);
  if i = length(UserTable)
  then exit
  else begin
    for j := i+1 to length(UserTable)-1 do UserTable[j-1] := UserTable[j];
    setlength(UserTable,length(UserTable)-1);
  end;
end; // function DeleteUserLabel

procedure MakeAsmBytes(    S : string; AddrMode : tAddrMode;
                       var AsmBytes : tAsmPackedRec);
var
  b1,b2,b3,b4 : byte;
  i           : integer;
begin
  for i := length(S) + 1 to 5 do S := S + chr(0);
  b1 := ord(S[1]) and $1F;
  b2 := ord(S[2]) and $1F;
  b3 := ord(S[3]) and $1F;
  b4 := ord(S[4]) and $1F;
  AsmBytes[1] := lo(b1 shl 3) + (b2 shr 2);
  AsmBytes[2] := lo(b2 shl 6) + lo(b3 shl 1) + (b4 shr 4);
  AsmBytes[3] := lo(b4 shl 4) + ord(AddrMode);
end; //

{Function to return the opcodes and operand format information contained
 in bytes 4 and 5 of AsmBytes, given the mnemonic and the address mode
 contained in bytes 1, 2 and 3.  If an address mode of amNone is supplied
 to this function, it will return without trying to match the address mode}
function  GetAsmBytes(var AsmBytes : tAsmPackedRec;
                      var ArrayIndex : integer) : boolean;
label
  Next;
begin
  ArrayIndex := 1;
  while (ArrayIndex <= AsmPackedArraySize) do
  begin
    if AsmBytes[1] <> AsmPA[ArrayIndex,1] then Goto Next;
    if AsmBytes[2] <> AsmPA[ArrayIndex,2] then Goto Next;
    if (AsmBytes[3] and $F0) <> (AsmPA[ArrayIndex,3] and $F0) then Goto Next;
    if tAddrMode(AsmBytes[3] and $0F) = amNone then break;
    if (AsmBytes[3] and $0F) = (AsmPA[ArrayIndex,3] and $0F) then break;
Next:
    inc(ArrayIndex);
  end;
  Result := ArrayIndex <= AsmPackedArraySize;
  if Result
  then AsmBytes := AsmPA[ArrayIndex]
  else ArrayIndex := 0;
end;

procedure ExtractAsmInfo(    ArrayIndex : integer;
                         var AddrMode   : tAddrMode;
//                         var NumModes   : integer;
                         var OpcodeStr,FormatStr : string);
var
  b3,b4,b5 : byte;
  OpStr    : string;
begin
  if (ArrayIndex < 1) or (ArrayIndex > AsmPackedArraySize) then exit;

  b3 := AsmPA[ArrayIndex][3];
  b4 := AsmPA[ArrayIndex][4];
  b5 := AsmPA[ArrayIndex][5];

  AddrMode := tAddrMode(b3 and $0F);
//  NumModes := (b4 and $30) shr 4;

  OpStr := IntToHex(b4,2);
  case (b5 shr 6) of
    0 : OpcodeStr := OpStr;
    1 : OpcodeStr := '18' + OpStr;
    2 : OpcodeStr := '1A' + OpStr;
    3 : OpcodeStr := 'CD' + OpStr;
  end;
  FormatStr := OperandFormats[b5 and $0F];
end; // procedure ExtractAsmInfo

procedure MakeAsmPA(i : integer; S : string);
var
  LList       : tStringList;
  b1,b2       : byte;
  AddrMode    : tAddrMode;
  Buf         : string;
begin
  LList := tStringList.Create;
  try
    LList.CommaText := S;

    AddrMode := amNone;
    if LList[3] = 'INH'   then AddrMode := amInherent;
    if LList[3] = 'IMM'   then AddrMode := amImmediate;
    if LList[3] = 'DIR'   then AddrMode := amDirect;
    if LList[3] = 'EXT'   then AddrMode := amExtended;
    if LList[3] = 'IND,X' then AddrMode := amIndirectX;
    if LList[3] = 'IND,Y' then AddrMode := amIndirectY;
    if LList[3] = 'REL'   then AddrMode := amRelative;
    if LList[3] = 'DRV'   then AddrMode := amDirective;

    MakeAsmBytes(LList[0],AddrMode,AsmPA[i]);

    b1 := 0;
    if LList[5] = ''
    then
      AsmPA[i,4] := StrToInt('$' + '0' + LList[4])
    else begin
      AsmPA[i,4] := StrToInt('$' + '0' + LList[5]);
      case LList[4][2] of
        '8' : b1 := 1;
        'A' : b1 := 2;
        'D' : b1 := 3;
      end;
    end;
    b1 := lo(b1 shl 6);

    Buf := LList[6] + LList[7] + LList[8];
    b2 := 0;
    if Buf = 'dd'     then b2 := 1;
    if Buf = 'ddmm'   then b2 := 2;
    if Buf = 'ddmmrr' then b2 := 3;
    if Buf = 'ff'     then b2 := 4;
    if Buf = 'ffmm'   then b2 := 5;
    if Buf = 'ffmmrr' then b2 := 6;
    if Buf = 'hhll'   then b2 := 7;
    if Buf = 'ii'     then b2 := 8;
    if Buf = 'jjkk'   then b2 := 9;
    if Buf = 'rr'     then b2 := 10;
    if Buf = 'bytes'  then b2 := 11;
    if Buf = 'ee'     then b2 := 12;
    if Buf = 'words'  then b2 := 13;
    if Buf = 'chars'  then b2 := 14;
    if Buf = 'space'  then b2 := 15;

    AsmPA[i,5] := b1 + b2;
  finally
    LList.Free;
  end;
end; // procedure MakeAsmPA

function OpBytesToString(    OperandText   : string;
                             BytesNotWords : boolean;
                         var DataStr       : string) : boolean;
var
  i,width : integer;
  SList   : tStringList;
begin
  Result  := false;
  DataStr := '';
  width   := 4;
  if BytesNotWords then width := 2;
  SList := tStringList.Create;
  try
    SList.CommaText := OperandText;
    for i := 0 to SList.Count - 1 do
    begin
      try
        DataStr := DataStr + IntToHex(StrToInt(SList[i]),width);
      except on EConvertError do exit;
      end;
    end;
    Result := true;
  finally
    if not Result then DataStr := '';
    SList.Free;
  end;
end; // function OpBytesToString

function Asm1Check(const CList : tStringList;
                   var Command : tCommand) : tCmdLine;
var
  A1,n1,n2  : word;
  AreasUsed : tMemAreaSet;
begin
  Result := clArgErr;                  // argument(s) not recognized
  if Prog.State in [psRunning,psTracing]
  then begin
    Result := clNotInRunTrace;
    exit;
  end;
  if MacroRunning
  then begin
    Result := clNotinMacros;
    exit;
  end;
  case Tokenize(CList,1,n1) of
    trByte,trWord :
      begin
        A1 := n1;
        if Tokenize(CList,2,n2) <> trNotExist then exit;
      end;
    trAmbiguousLabel : begin
                         Result := clAmbiguousLabel;
                         exit;
                       end
    else exit;
  end; // case

{Check for address range exclusions when writing to MCU controlled memory}
  if not IsLocal
  then begin
    if not CheckAreas(A1,A1,[maRam,maEeprom],AreasUsed,Result)
    then exit;
  end;
  Prog.StartAddr := A1;
  case IsLocal of
    true  : Command := cmdAssembleLocal;
    false : Command := cmdAssemble;
  end;
  Result := clOK;
end; // function Asm1Check

{Asm2Check is the main assembling routine. It does the conversion from a
 source line to machine code.}
function Asm2Check(const CList     : tStringList;
                       CmdText     : string;
                       LastCommand : tCommand;
                   var Command     : tCommand) : tCmdLine;
var
  b1,b2                       : byte;
  n1,PC,DummyWord,NextPC      : word;
  i,j,k,SCPos,I1,ArrayIndex   : integer;
  AddrMode                    : tAddrMode;
  Within1Quotes,within2Quotes,HasHash,NewLabel,Ambiguous : boolean;
  LabelText,MnemonicText,OperandText,Buf,FormatStr,OpcodeStr,
  QuoteText,PartStr,TempStr   : string;
  AsmBytes                    : tAsmPackedRec;
  AreasUsed                   : tMemAreaSet;
  TokenResult                 : tTokenResult;
Label
  Next;
begin
  Result       := clOK;
  Command      := LastCommand;
  AsmErr       := aeNone;
  PC           := Prog.StartAddr;
  LabelText    := '';
  MnemonicText := '';
  OperandText  := '';
  DataStr      := '';
  b1 := $FF; b2 := $FF;
  NewLabel     := false;

{If previous command was not cmdAssembling then exit}
  if not (LastCommand in [CmdAssembling,CmdAssemblingLocal])
  then begin
    Command := cmdNone;
    exit;
  end;

{Typing Return at the > prompt causes the ASM command to terminate}
  i := pos('>',CmdText);
  if (i = 0) or (i = length(CmdText))
  then begin
    Command := cmdEnd;
    Result  := clTerminate;
    exit;
  end;

  try
{Remove any comment}
    SCPos := pos(';',CmdText);
    if SCPos > 0
    then delete(CmdText,SCPos,length(CmdText) - SCPos + 1);
{Parse the line into its elements}
    i := i + 1;
{If the first character is not a space, then the element must be a label}
    if copy(CmdText,i,1) <> ' '
    then begin
      while (i <= length(CmdText)) and (CmdText[i] <> ' ') do
      begin
        LabelText := LabelText + CmdText[i];
        inc(i);
      end;
{If last character is a colon, remove it}
      if copy(LabelText,length(LabelText),1) = ':'
      then delete(LabelText,length(LabelText),1);
{Check that only allowed chars are in the label}
      if not LabelCheck(LabelText)
      then begin
        AsmErr := aeIllegalCharInLabel;
        exit;
      end;
{Labels identical with mnemonics are not allowed}
      MakeAsmBytes(LabelText,amNone,AsmBytes);
      if GetAsmBytes(AsmBytes,ArrayIndex)
      then begin
        AsmErr := aeIllegalLabel;
        exit;
      end;
{Is the label already known to JBug11 ?}
      if LookupLabelValue(LabelText,n1,Ambiguous)
      then begin
        if Ambiguous
        then NewLabel := true
        else begin
          LookupLabelName(n1,TempStr);
          if (TempStr <> LabelText)
          then NewLabel := true
          else begin
            if (n1 <> PC)
            then begin
              AsmErr := aeDuplicateLabel;
              exit;
            end;
          end;  
        end;
      end
{Must be new label - add it to the user's list of labels}
      else NewLabel := true;
    end; // if copy(CmdText,1,1) <> ' ' then..

    while (i <= length(CmdText)) and (CmdText[i] = ' ') do inc(i);

{Next element should be an Instruction or a Directive}
    while (i <= length(CmdText)) and (CmdText[i] <> ' ') do
    begin
      MnemonicText := MnemonicText + upcase(CmdText[i]);
      inc(i);
    end;
{If no mnemonic, then error}
    if length(MnemonicText) = 0
    then begin
      AsmErr := aeNoMnemonic;
      exit;
    end;
{Is this a recognized mnemonic? Use MakeAsmBytes with AddrMode = amNone, this
 is a special case which returns the first listed occurrence of AsmBytes
 without matching the value of AddrMode}
    MakeAsmBytes(MnemonicText,amNone,AsmBytes);
    if (not GetAsmBytes(AsmBytes,ArrayIndex)) or (length(MnemonicText) > 5)
    then begin
      AsmErr := aeUnrecogMnemonic;
      exit;
    end;

{If a new label was previously detected, then either add it to the list with
 the value of the program counter, or wait until later if mnemonic is the EQU
 directive}
    if NewLabel
    then if MnemonicText <> 'EQU' then AddUserLabel(LabelText,PC);

    while (i <= length(CmdText)) and (CmdText[i] = ' ') do inc(i);

{Next element, if there is one, will be an Operand.  This is copied to
 'OperandText' by the following rather complicated procedure which is
 necessary to take account of quoted text passages which have to be
 recognized even where they contain spaces}
    if i <= length(CmdText)
    then begin
      QuoteText     := '';
      Within1Quotes := false;
      Within2Quotes := false;
      while (i <= length(CmdText)) do
      begin
{If an opening or closing single quote is encountered:}
        if CmdText[i] = ''''
        then begin
          if Within1Quotes
          then begin
            Within1Quotes := false;
            for j := 1 to length(QuoteText) do
              OperandText := OperandText + '$' + IntToHex(ord(QuoteText[j]),2);
            QuoteText := '';
          end
          else begin
            if Within2Quotes
            then QuoteText := QuoteText + CmdText[i]
            else Within1Quotes := true;
          end;
          Goto Next;
        end;
{If an opening or closing double quote is encountered:}
        if CmdText[i] ='"'
        then begin
          if Within2Quotes
          then begin
            Within2Quotes := false;
            for j := 1 to length(QuoteText) do
              OperandText := OperandText + '$' + IntToHex(ord(QuoteText[j]),2);
            QuoteText := '';
          end
          else begin
            if Within1Quotes
            then QuoteText := QuoteText + CmdText[i]
            else Within2Quotes := true;
          end; // if Within2Quotes then.. else..
          Goto Next;
        end; // if CmdText[i] ='"' then..

{Copy the character in CmdText to OperandText or QuoteText as appropriate}
        if Within1Quotes or Within2Quotes
        then QuoteText := QuoteText + CmdText[i]
        else begin
          if CmdText[i] =' '
          then break
          else OperandText := OperandText + CmdText[i];
        end;  
Next:   inc(i);
      end; // while (i < length(CmdText)) do
{Quoted strings should have terminated:}
      if Within1Quotes or Within2Quotes
      then begin
        AsmErr := aeUnterminatedQuote;
        exit;
      end
    end; // if i <= length(CmdText) then.. else..

{By now i should be at one beyond the end of the command line}
    if i <> length(CmdText) + 1
    then begin
      AsmErr := aeIntErr1;
      exit;
    end;

{If OperandText begins with a comma, add a zero at the beginning}
    if copy(OperandText,1,1) = ','
    then insert('0',OperandText,1);

{Parse OperandText and check it for correct syntax}
{Determine type of addressing.  Note that the earlier call to MakeAsmBytes with
 AddrMode = amNone, followed by a call to GetAsmBytes, will have set 'Index',
 and if the address mode returned by that last call to GetAsmBytes is either
 amInherent or amRelative, then there is no need to try and determine the
 addressing mode by inspection of the operands, as instructions with these
 two address modes do not have more than one mode}
    ExtractAsmInfo(ArrayIndex,AddrMode,OpcodeStr,FormatStr);
    if AddrMode in [amNone,amInherent,amRelative,amDirective]
    then begin end
    else begin
      if length(OperandText) = 0
      then begin
        AsmErr := aeNoOperand;
        exit;
      end;
      if copy(OperandText,1,1) = '#'
      then AddrMode := amImmediate
      else begin
        if pos(',X',uppercase(OperandText)) > 0
        then AddrMode := amIndirectX
        else begin
          if pos(',Y',uppercase(OperandText)) > 0
          then AddrMode := amIndirectY
          else begin
            GetOperandPart(OperandText,1,PartStr);
            case ExtConversion(PartStr,DummyWord) of
              trByte   : AddrMode := amDirect;
              trWord   : AddrMode := amExtended;
              trNotRecognized :
                begin
                  if not LookupLabelValue(PartStr,n1,Ambiguous) or Ambiguous
                  then begin
                    if PartStr = '*'
                    then n1 := PC
                    else begin
                      AsmErr := aeLabelNotDefined;
                      exit;
                    end;
                  end;
                  if n1 <= $FF
                  then AddrMode := amDirect
                  else AddrMode := amExtended;
                end // trNotRecognized
            else begin AsmErr := aeIntErr4; exit; end;
            end; // case ExtConversion(PartStr,DummyWord) of
          end; // if pos(',Y',OperandText) > 0 then.. else..
        end; // if pos(',X',OperandText) > 0 then.. else..
      end; // if copy(OperandText,1,1) = '#' then.. else..
    end; // if AddrMode in [amNone,amInherent,amRelative] then.. else..

{Look up the operand format string, and the opcode bytes}
    MakeAsmBytes(MnemonicText,AddrMode,AsmBytes);
{$B+}
    if (AddrMode = amDirect) and not GetAsmBytes(AsmBytes,ArrayIndex)
    then begin
      MakeAsmBytes(MnemonicText,amExtended,AsmBytes);
      if not GetAsmBytes(AsmBytes,ArrayIndex)
      then begin
        AsmErr := aeIntErr5;
        exit;
      end;
    end;
{$B-}
    ExtractAsmInfo(ArrayIndex,AddrMode,OpcodeStr,FormatStr);

    if AddrMode <> amDirective then DataStr := OpcodeStr;
{Using the FormatStr as a guide, 'fill in the blanks' or generate an
 appropriate error message}
    j := 1;
    while (j <= 4) and (FormatStr[j] <> ' ') do
    begin
      Ambiguous := false;
      HasHash   := false;
      GetOperandPart(OperandText,j,Buf);
      if copy(Buf,1,1) = '#'
      then begin
        HasHash := true;
        delete(Buf,1,1);
      end;
      TokenResult := ExtConversion(Buf,n1);
      if (TokenResult = trNotRecognized) and
         (FormatStr[j] <> 'X')
//         and (AddrMode <> amDirective)
      then begin
        if not LookupLabelValue(Buf,n1,Ambiguous) or Ambiguous
        then begin
          if Buf = '*'
          then n1 := PC
          else begin
            AsmErr := aeLabelNotDefined;
            exit;
          end;
        end;
      end;

      if (n1 > $FF) and (FormatStr[j] in ['D','F','I','M'])
      then begin
        AsmErr := aeOperandOutOfRange;
        exit;
      end;
      if (FormatStr[j] in ['I','J']) and not HasHash
      then begin
        AsmErr := aeIntErr2;
        exit;
      end;

      case FormatStr[j] of
        'D' : begin
                b1 := lo(n1);
              end;
        'F' : begin
                b1 := lo(n1);
              end;
        'X' : if ((AddrMode = amIndirectX) and (Buf <> 'X')) or
                 ((AddrMode = amIndirectY) and (Buf <> 'Y'))
              then begin
                AsmErr := aeOperandError;
                exit;
              end;
        'H' : begin
                b1 := hi(n1); b2 := lo(n1);
              end;
        'I' : begin
                if n1 > $FF
                then begin
                  AsmErr := aeOperandOutOfRange;
                  exit;
                end;
                b1 := lo(n1);
              end;
        'J' : begin
                b1 := hi(n1); b2 := lo(n1);
              end;
        'K' : begin
                b1 := b2;
              end;
        'L' : begin
                b1 := b2;
              end;
        'M' : begin
                if n1 > $FF
                then begin
                  AsmErr := aeOperandOutOfRange;
                  exit;
                end;
                b1 := lo(n1);
              end;
        'R' : begin
{Determine what the PC of the following instruction would be}
                NextPC := word(PC + (length(DataStr) div 2) + 1);
{Check that the branch is within range}
                if (n1 < (NextPC - $80)) or (n1 > (NextPC + $7F))
                then begin
                  AsmErr := aeBranchOutOfRange;
                  exit;
                end;
{Develop the relative offset byte}                
                I1 := n1 - NextPC;
                b1 := lo(I1);
              end; // 'R':
        'B' : begin
                if not OpBytesToString(OperandText,true,DataStr)
                then begin
                  AsmErr := aeOperandError;
                  exit;
                end;
              end;
        'E' : begin
                if NewLabel
                then AddUserLabel(LabelText,n1)
                else begin
                  AsmErr := aeLabelMissing;
                  exit;
                end;
              end;
        'W' : begin
                if not OpBytesToString(OperandText,false,DataStr)
                then begin
                  AsmErr := aeOperandError;
                  exit;
                end;
              end;
        'C' : begin
                DataStr := '';
                for k := 1 to length(Buf) do
                begin
                  if (Buf[k] <> '$') and (Buf[k] in HexChars)
                  then DataStr := DataStr + Buf[k];
                end;
              end;
        'S' : begin
                DataStr := '';
                if PC + n1 > $FFFF
                then begin
                  AsmErr := aeOperandOutOfRange;
                  exit;
                end;
                PC := PC + n1;
                Prog.StartAddr := PC;
              end;
      end; // case
      
      if (AddrMode <> amDirective) and (FormatStr[j] <> 'X')
      then DataStr := DataStr + IntToHex(b1,2);

      inc(j);
    end; // while (j <= 3) and (FormatStr[j] <> ' ') do

    Prog.EndAddr := PC + (length(DataStr) div 2) - 1;
{Check memory ranges }
    if (LastCommand = cmdAssembling) and not CheckAreas(Prog.StartAddr,
                                             Prog.EndAddr,[maRam,maEeprom],
                                             AreasUsed,Result)
    then exit;

  finally
    MainForm.StatusBar1.Panels[1].Text := AsmErrMsgs[AsmErr];
    if AsmErr <> aeNone
    then begin
      MainForm.CmdEdit.SelStart := i-1;
      Result  := clAsmError;
      DataStr := '';
      Beep;
    end;
  end;
end; // function Asm2Check

procedure AsmMemory;
var
  i,j : integer;
begin
  with Prog do
  begin
    if length(DataStr) > 0
    then begin
      i := 0;
      j := 1;
      while j < length(DataStr) do
      begin
        MemArray[StartAddr + i] := StrToInt('$' + copy(DataStr,j,2));
        i := i + 1;
        j := j + 2;
      end;
    end;
  end;
end; // procedure AsmMemory

procedure AsmLocal1(var Command : tCommand);
var
  Buf : string;
begin
  BPROTChecked := false;
  with MainForm.CmdEdit,Prog do
  begin
    Text := format('%6.4x',[StartAddr]) + ' >';
    if LookupLabelName(StartAddr,Buf)
    then Text := Text + Buf + ' ';
    Command  := cmdAssemblingLocal;
    SelStart := length(Text);          // move cursor to end of line
  end; // with MainForm.CmdEdit,Prog do
end; // procedure AsmLocal1

procedure AsmLocal2(var Command : tCommand);
begin
  with Prog do
  begin
    if length(DataStr) > 0
    then begin
      AsmMemory;
      StartAddr := EndAddr + 1;
    end;
    if AsmErr = aeNone then AsmLocal1(Command);
  end;
end; // procedure AsmLocal2

procedure Asm1(var Command : tCommand);
begin
  AsmLocal1(Command);
  Command := cmdAssembling;
end; // procedure Asm1

procedure Asm2(var Command : tCommand);
begin
  with Prog do
  begin
    if length(DataStr) > 0
    then begin
      AsmMemory;
      WriteMCUMem(StartAddr,EndAddr,[maRam,maEeprom],false,false);
      StartAddr := EndAddr + 1;
    end;
    if AsmErr = aeNone then Asm1(Command);
  end;
end; // procedure Asm2

{ExtConversion will take a text string of number characters and attempt to
 convert it into a word or byte.  If it succeeds, then its result will be
 trWord or trByte.
 It can cope with Hex (numbers preceded by $), decimal, and binary (numbers
 preceded by %)}
function  ExtConversion(    S : string;
                        var W : word) : tTokenResult;
var
  i : integer;
  L : longword;
begin
  Result := trNotRecognized;
//  if (NumBytes < 1) or (NumBytes > 2) then exit;
  if (copy(S,1,1) = '%') and (length(S) = 9)
  then begin
    for i := 2 to 9 do if (S[i] <> '0') and (S[i] <> '1') then exit;
    L := 0;
    for i := 0 to 7 do L := L + ((ord(S[9-i]) - ord('0')) shl i);
  end
  else begin
    try
      L := StrToInt(S);
    except on EConvertError do exit;
    end;
  end;
  if L > $FFFF then exit;
  W := L;
  if L > $FF
  then Result := trWord
  else Result := trByte;
end; // function  ExtConversion

procedure GetOperandPart(    OperandText : string;
                             PartNum     : integer;
                         var PartStr     : string);
var
  i,PartCount                 : integer;
  Within1Quotes,Within2Quotes : boolean;
begin
  Within1Quotes := false;
  Within2Quotes := false;
  PartStr       := '';
  PartCount     := 0;
  i := 0;
  OperandText := OperandText + ',';   // add a final terminating comma
  while (i < length(OperandText)) do
  begin
    inc(i);
    PartStr := PartStr + OperandText[i];
{If an opening or closing single quote is encountered:}
    if OperandText[i] =''''
    then begin
      if Within1Quotes
      then Within1Quotes := false
      else begin
        if Within2Quotes
        then begin end
        else Within1Quotes := true;
      end; // if Within1Quotes then.. else..
      continue;
    end;

{If an opening or closing double quote is encountered:}
    if OperandText[i] ='"'
    then begin
      if Within2Quotes
      then Within2Quotes := false
      else begin
        if Within1Quotes
        then begin end
        else Within2Quotes := true;
      end; // if Within2Quotes then.. else..
      continue;
    end; // if CmdText[i] ='"' then..

{If a comma character is encountered, this either signals the start of a new
 part of OperandText, or is part of a quoted string if within quotation marks}
    if OperandText[i] =','
    then begin
      if Within1Quotes or Within2Quotes
      then begin end
      else begin
        PartCount := PartCount + 1;
        delete(PartStr,length(PartStr),1);  // delete the comma
        if PartCount = PartNum
        then break
        else PartStr := '';
      end; // if Within1Quotes or Within2Quotes then.. else..
    end; // if CmdText[i] =',' then..
  end; // while (i < length(OperandText)) do
{If PartStr is of zero length}
  if length(PartStr) = 0 then exit;
end; // function  GetOperandPart

end.

