unit RunTraceUnit;

interface

uses Forms, Classes, SysUtils, Dialogs, Controls, TypeDefs;

function TraceCheck(const CList   : tStringList;
                    var   Comment : tCmdLineComment;
                    var   Command : tCommand) : boolean;
function GoCheck(const CList   : tStringList;
                 var   Comment : tCmdLineComment;
                 var   Command : tCommand) : boolean;
function StopCheck(const CList   : tStringList;
                   var   Comment : tCmdLineComment;
                   var   Command : tCommand) : boolean;

procedure DiscoverNext(var BPRecord   : tBPRecord;
                           StackFrame : tStackFrame;
                           IsStepOver : boolean);

procedure Trace;
procedure Go;
procedure HandleSWI;
procedure Stop;

implementation

uses Main, CmdUnit, UnAsmUnit, MemRW, LookupUnit, MiscUnit, CheckUnit, BPUnit,
     RegRW, Comms1, AddToWatchUnit;

var
  VectorsSwapped      : boolean = false;
  VectorsRead         : boolean = false;
  StartAddr,BreakAddr : word;
  IsStepOver          : boolean;
  StartAddrSpecified  : boolean;    // used to control display of insts & regs
  Repeats             : integer;
  StartRecord         : tBPRecord;
  BreakRecord         : tBPRecord;
  ATArray             : tATArray;

function TraceCheck(const CList   : tStringList;
                    var   Comment : tCmdLineComment;
                    var   Command : tCommand) : boolean;
var
  AreasUsed  : tMemAreaSet;
  ArgType    : tArgType;
  Buf        : string;
  Val        : longword;
begin
  Result    := false;
  Comment   := clcArgErr;
  StartAddr := 0;

  if not NumArgsOK(CList,0,1,Comment) then exit;

  if not StateIsAllowed([psStopped,psRunStop,psTraceStop],Comment) then exit;

  Buf := uppercase(CList[0]);
  if (Buf = 'T') or (Buf = 'O')
  then Repeats := 1
  else begin
    if length(CList[0]) > 1 then delete(Buf,1,1);
    try
      Repeats := StrToInt(Buf);
    except on EConvertError do begin Comment := clcCmdErr; exit; end;
    end; // try.. except on EConvertError do
  end; // if (Buf = 'T') or (Buf = 'O') then.. else..

  IsStepOver := copy(Buf,1,1) = 'O';

  if FoundArgOK(CList,1,[atNotExist,atWord,atLabel,atAsterisk],ArgType,Val)
  then begin
    case ArgType of
      atNotExist,atAsterisk : StartAddr := Prog.TGAddr;
      atWord,atLabel        : StartAddr := Val;
    end;
  end else exit;

  StartAddrSpecified := ArgType in [atWord,atLabel];

  if not CheckAreas(StartAddr,StartAddr,[maRam,maExtRam,maEeprom],
                    MemMap,AreasUsed,Comment)
  then exit;

  if not VectorsRead and not ProjData.NoVectorSubstitution
  then begin
{Read the jump vector locations, and store the results in UserSwiVector and
 UserIllopVector}
    UserSwiVector   := ReadWord(MapData.swi_jmp);
    UserIllopVector := ReadWord(MapData.illop_jmp);
    VectorsRead     := true;
  end;

  MakeBPRecord(StartAddr,bpStart,StartRecord);
  with StartRecord,MapData do
  begin
    if (InstType = itIllegalOpcode) and (UserIllopVector = talker_start)
    then begin Comment := clcInvalidStartCode; exit; end;
    if (InstType = itSWI) and (UserSWIVector = swi_srv)
    then begin Comment := clcInvalidStartSWI; exit; end;
  end;

  with StartRecord do
  begin
    if ((SeqAddr < $FFFF) and (GetMemArea(SeqAddr,MemMap) = maEeprom)) or
       ((AltAddr < $FFFF) and (GetMemArea(AltAddr,MemMap) = maEeprom))
    then begin
      if not BprotOK then BprotOK := BprotWarningDlg;
      if not BprotOK
      then begin
        Comment := clcBPROTMayDisallowWrite;
        exit;
      end;
    end;
  end;

  Result    := true;
  Comment   := clcOK;
  if IsStepOver then Command := cmdStepOver else Command := cmdTrace;
end; // function TraceCheck

function GoCheck(const CList   : tStringList;
                 var   Comment : tCmdLineComment;
                 var   Command : tCommand) : boolean;
var
  AreasUsed   : tMemAreaSet;
  ArgType     : tArgType;
  i,j         : integer;
  BPPtr       : tBPRecordPtr;
  Val         : longword;
  BreakType   : tBPType;
  Buf         : string;
  RegAddr     : word;
begin
  Result             := false;
  Comment            := clcArgErr;
  StartAddr          := 0;
  BreakAddr          := 0;
  BreakType          := bpNone;
  Buf                := '';
  ATArray            := nil;

  if not NumArgsOK(CList,0,6,Comment) then exit;

  if not StateIsAllowed([psStopped,psRunStop,psTraceStop],Comment) then exit;

{Has a starting address been specified}
  if FoundArgOK(CList,1,[atNotExist,atWord,atLabel,atAsterisk],
                ArgType,Val)
  then begin
    case ArgType of
      atNotExist,atAsterisk : StartAddr := Prog.TGAddr;
      atWord,atLabel        : begin
                                StartAddr := Val;
                                StartAddrSpecified := true;
                              end;
    end;
  end else exit;

  if not CheckAreas(StartAddr,StartAddr,
                    [maRam,maRom,maEeprom,maEprom,maExtRam,maByteWrite,maPageWrite],
                    MemMap,AreasUsed,Comment) then exit;

  if not VectorsRead and not ProjData.NoVectorSubstitution
  then begin
{Read the jump vector locations, and store the results in UserSwiVector and
 UserIllopVector}
    UserSwiVector   := ReadWord(MapData.swi_jmp);
    UserIllopVector := ReadWord(MapData.illop_jmp);
    VectorsRead     := true;
  end;

  MakeBPRecord(StartAddr,bpStart,StartRecord);
  with StartRecord,MapData do
  begin
    if (InstType = itIllegalOpcode) and (UserIllopVector = talker_start)
    then begin Comment := clcInvalidStartCode; exit; end;
    if (InstType = itSWI) and (UserSWIVector = swi_srv)
    then begin Comment := clcInvalidStartSWI; exit; end;
  end;

{Any subsequent arguments must be breakpoints}
  if CList.Count > 2
  then begin
    setlength(ATArray,CList.Count - 2);
    ClearATArray(ATArray);
  end;
  for i := 2 to CList.Count - 1 do
  begin
    if FoundArgOK(CList,i,[atWord,atWordX,atLabel],ArgType,Val)
    then begin
      case ArgType of
        atWord,atLabel :
          begin
            BreakAddr := Val;
            BreakType := bpTransient;
          end;
        atWordX :
          begin
            BreakAddr := Val;
            BreakType := bpFixed;
          end;
      end; // case ArgType of
    end; // if FoundArgOK(CList,i,[atWord,atWordX,atLabel],ArgType,Val) then..

{Check proposed memory area for this breakpoint}
    if not CheckAreas(BreakAddr,BreakAddr,[maRam,maExtRam,maEeprom],
                      MemMap,AreasUsed,Comment) then exit;

{Check that BPROT is OK, if necessary}
    if (maEeprom in AreasUsed) and
       (LookupCtrlRegAddr('BPROT',RegAddr,j))and
       ProjData.WarnOnBprot
    then begin
      if (ReadByte(RegAddr) and $0F) > 0
      then begin
        Comment := clcBPROTMayDisallowWrite;
        exit;
      end;
    end;

{Check that this breakpoint is not already in the list}
    if RetrieveBP(BreakAddr,[bpTransient..bpPass],BPPtr)
    then begin
      Comment := clcDuplicateBP;
      exit;
    end;

{If OK add to the address transfer list}
    if ATAddrExists(BreakAddr,ATArray)
    then begin
      Comment := clcDuplicateBP;
      exit;
    end;
    ATArray[i-2].Addr   := BreakAddr;
    ATArray[i-2].BPType := BreakType;
  end; // for i := 2 to CList.Count - 1 do
  Result    := true;
  Comment   := clcOK;
  Command   := cmdGo;
end; // function GoCheck

function StopCheck(const CList   : tStringList;
                   var   Comment : tCmdLineComment;
                   var   Command : tCommand) : boolean;
begin
  Result    := false;
  Comment   := clcArgErr;

  if not NumArgsOK(CList,0,0,Comment) then exit;

  if not StateIsAllowed([psStopped,psRun,psTrace,psRunStop,psTraceStop],Comment)
  then exit;

  Result    := true;
  Comment   := clcOK;
  Command   := cmdStop;
end; // function StopCheck

{****************************************************************************}

procedure SwapTVectorsIn;
begin
  if VectorsSwapped then exit;
  with MapData,ProjData do
  begin
{If vector substitution is allowed, then overwrite the vector locations with
 the talker service routine entry points}
    if not NoVectorSubstitution
    then begin
      WriteWord(swi_jmp,swi_srv,true);
      if (illop_jmp > 0)       // i.e. illop_jmp is defined in the map file
      then WriteWord(illop_jmp,talker_start,true);
      VectorsSwapped := true;
    end;
  end;
end; // procedure SwapTVectorsIn

procedure SwapTVectorsOut;
begin
  if not VectorsSwapped then exit;
  with MapData,ProjData do
  begin
    if not NoVectorSubstitution
    then begin
      WriteWord(swi_jmp,UserSwiVector,true);
      if (illop_jmp > 0)       // i.e. illop_jmp is defined in the map file
      then WriteWord(illop_jmp,UserIllopVector,true);
      VectorsSwapped := false;
    end;
  end;
end; // procedure SwapTVectorsOut

{If an SWI is encountered which has been placed by the user in MCU code,
 then it is necessary to run an emulation of the SWI operation, otherwise
 the OnRxChar breakpoint handler will endlessly interrupt processing}
procedure EmulateSWI(    SWIAddr     : word;
                     var LocalFrame  : tStackFrame;
                         StoppedAtBP : boolean);
var
  ReturnAddr : word;
  DummyAddr  : word;
begin
{If not stopped at a breakpoint, e.g. Going at an SWI without having arrived
 there during debugging, then it is necessary to ensure that the CPU is
 idling at SWIidle}
  if not StoppedAtBP then ServiceSWI(LocalFrame,DummyAddr);

  with LocalFrame do
  begin
{Copy the current stack to a new frame, 9 addresses lower on the stack, and
 set the MCU running temporarily - CPU will run at the SWI idle loop}
    SP := SP - 9;
    WriteMCUReg(LocalFrame);

{Increment the BP address noted when the SWI was signalled, so that on return
 the PC will point to the instruction following the SWI}
    ReturnAddr := SWIAddr + 1;

{Write the PC value in the stack frame to be used when the RTI is encountered}
    WriteWord(SP + 8,ReturnAddr,true);

{Set the MCU running at the SWI service routine entry point}
    PC  := UserSWIVector;
    CCR := CCR or $10;                        // set the I bit in the CCR
  end;
end; // procedure EmulateSWI

function HaltQuery(BPRecord : tBPRecord; State : tProgState) : boolean;
var
  Triggered : boolean;
begin
  with BPRecord,ProjData do
  begin
    Triggered := PassCount >= Trigger;
    case State of
      psRun :
        case BPType of
          bpTemp              : Result := false;
          bpTransient,bpFixed : Result := true;
          bpPass              : Result := Triggered
          else Result := true;
        end;
      psTrace : Result := IsNextTraceBP or (not NotInStepOver)
      else Result := true;
    end; // case State of
  end; // with BPRecord,ProjData do
end; // function HaltQuery

{Procedure to generate the next breakpoint address, given a breakpoint record
 at the current instruction, plus a stack frame to supply IX, IY and the CCR.
 Special processing is necessary for the instructions BRCLR and BRSET, this is
 done by an additional read of MCU memory followed by a simulation of the
 CPU action. }
procedure AnalyseBranchAddr(var BPRecord       : tBPRecord;
                                StackFrame     : tStackFrame);
var
  Overflow,Negative,Zero,Carry : boolean;
  DataAddr                     : word;
  Mask,DataByte                : byte;
begin
{Set up the CCR flags}
  with StackFrame do
  begin
    Negative  := (CCR and $08) > 0;
    Zero      := (CCR and $04) > 0;
    Overflow  := (CCR and $02) > 0;
    Carry     := (CCR and $01) > 0;
  end;

  with BPRecord do
  begin
{Make AltAddr the default}
    NextAddr := AltAddr;

{At branch instructions, the next breakpoint will depend upon the
 instruction at the present address and the values in the CCR }
    if ((Inst = 'BHS') or (Inst = 'BCC')) and not Carry          then exit;
    if ((Inst = 'BLO') or (Inst = 'BCS')) and Carry              then exit;
    if (Inst = 'BEQ')  and Zero                                  then exit;
    if (Inst = 'BGE')  and not (Negative xor Overflow)           then exit;
    if (Inst = 'BGT')  and not (Zero or (Negative xor Overflow)) then exit;
    if (Inst = 'BHI')  and not (Carry or Zero)                   then exit;
    if (Inst = 'BLE')  and (Zero or (Negative xor Overflow))     then exit;
    if (Inst = 'BLS')  and (Carry or Zero)                       then exit;
    if (Inst = 'BLT')  and (Negative xor Overflow)               then exit;
    if (Inst = 'BMI')  and Negative                              then exit;
    if (Inst = 'BNE')  and not Zero                              then exit;
    if (Inst = 'BPL')  and not Negative                          then exit;
    if (Inst = 'BVC')  and not Overflow                          then exit;
    if (Inst = 'BVS')  and Overflow                              then exit;

    if (Inst = 'BRA')                                            then exit;
    if (Inst = 'BRN')  then begin end;

    if ((Inst = 'BRCLR') or (Inst = 'BRSET'))
    then begin
      Mask := Opd2;
      case OpcodeDatumPtr^.AddrMode of
        amDIR : DataAddr := Opd1;
        amIX  : DataAddr := StackFrame.IX + Opd1;
        amIY :  DataAddr := StackFrame.IY + Opd1
        else DataAddr := 0;
      end; // case AddrMode of
      DataByte := ReadByte(DataAddr);
      if Inst = 'BRSET' then DataByte := $FF - DataByte;
      if (DataByte and Mask) = 0                                 then exit;
    end;

    NextAddr := SeqAddr;
  end; // with BPRecord do
end; // procedure AnalyseBranchAddr

{Procedure to discover the address of the next instruction to be executed}
procedure DiscoverNext(var BPRecord   : tBPRecord;
                           StackFrame : tStackFrame;
                           IsStepOver : boolean);
begin
{  tInstructionType = (itNone,itIllegalOpcode,itSimple,itBranch,itBSR,itJSR,
                      itJump,itJumpIX,itJumpIY,itRTI,itRTS,itSWI) }
  with BPRecord do
  begin
    case InstType of
      itNone   : NextAddr := $FFFF;
      itIllegalOpcode :
                 begin
                   NextAddr := UserIllopVector;
                   SeqAddr  := $FFFF;
                 end;
      itSimple : NextAddr := SeqAddr;
      itBranch : AnalyseBranchAddr(BPRecord,StackFrame);
      itBSR,itJSR :
                 if IsStepOver
                 then NextAddr := SeqAddr
                 else NextAddr := AltAddr;
      itJump   : NextAddr := AltAddr;
      itJumpIX : NextAddr := StackFrame.IX;
      itJumpIY : NextAddr := StackFrame.IY;
      itRTI    : NextAddr := ReadWord(StackFrame.SP + 8);
      itRTS    : NextAddr := ReadWord(StackFrame.SP + 1);
      itSWI    : if IsStepOver
                 then NextAddr := SeqAddr
                 else NextAddr := UserSWIVector
      else NextAddr := $FFFF;
    end; // case InstType of

{Mod added 2006-04-18 to improve display in the BP display window}
    if InstType in [itJumpIX,itJumpIY,itRTI,itRTS,itSWI,itIllegalOpcode]
    then AltAddr := NextAddr;

  end; // with BPRecord do
end; // procedure DiscoverNext

procedure MakeAndAddTempBP(BPRecord     : tBPRecord;
                           Frame        : tStackFrame);
var
  NewBPRecord : tBPRecord;
  BPPtr       : tBPRecordPtr;
begin
  DiscoverNext(BPRecord,Frame,false);
  with BPRecord do
  begin
    if NextAddr = $FFFF then exit;
    if not RetrieveBP(NextAddr,[bpTemp..bpPass],BPPtr)
    then begin
      MakeBPRecord(NextAddr,bpTemp,NewBPRecord);
      AddBPToList(NewBPRecord);
    end
    else begin
      BPPtr^.ActingAsTempBP := true;
    end;
  end;
end; // procedure MakeAndAddTempBP

procedure MakeAndAddTransientBP(BPRecord   : tBPRecord;
                                Frame      : tStackFrame;
                                IsStepOver : boolean);
var
  NewBPRecord : tBPRecord;
  BPPtr       : tBPRecordPtr;
begin
  DiscoverNext(BPRecord,Frame,IsStepOver);
  with BPRecord do
  begin
    if NextAddr = $FFFF then exit;
    if not RetrieveBP(NextAddr,[bpTemp..bpPass],BPPtr)
    then begin
      MakeBPRecord(NextAddr,bpTransient,NewBPRecord);
      NewBPRecord.IsNextTraceBP := true;
      AddBPToList(NewBPRecord);
    end
    else begin
      BPPtr^.IsNextTraceBP := true;
    end;
  end;
end; // procedure MakeAndAddTransientBP

{*****************************************************************************}

procedure Trace;
var
  i                : integer;
  LocalFrame       : tStackFrame;
  Flag,StoppedAtBP : boolean;
begin
  SwapTVectorsIn;
  if not InitialReadRegs
  then begin
    ReadMCUReg(LocalFrame);
    Prog.StackFrame := LocalFrame;
  end;
  for i := 1 to Repeats do
  begin
    DeleteBPByType([bpStart]);
{Remove any Transient breakpoint at the starting address}
    DeleteBPAtAddr(StartAddr,[bpTransient]);
//    RemoveTransientSetByTrace;
    RemoveTraceRequests;
    StoppedAtBP := StartAddr = Prog.BRAddr;
    Prog.BRAddr := $FFFF;

    MakeAndAddTransientBP(StartRecord,Prog.StackFrame,IsStepOver);
    AddBPtoList(StartRecord);
    UpdateBPWindow;
    SetBreaksX([bpTemp..bpPass],StartRecord.Addr);
    CommsForm.ComPort.OnRxChar := CommsForm.ComPortRxChar1;
    Prog.State := psTrace;

    MainForm.PCLbl.Caption := IntToHex(StartAddr,4);
    LocalFrame := Prog.StackFrame;

    if ProjData.ShowOnceOnly
    then Flag := StartAddrSpecified
    else Flag := true;
    if Flag
    then begin
      if ProjData.ShowStartInst
      then MainForm.OWEdit.Lines.Add('Start Inst: ' + StartRecord.OWStr);
      if ProjData.ShowStartRegs
      then MainForm.OWEdit.Lines.Add('Start Regs: ' + FrameToString);
    end;
    StartAddrSpecified := false;

    if StartRecord.InstType = itSWI
    then begin
      EmulateSWI(StartAddr,LocalFrame,StoppedAtBP);
      LocalFrame.PC := UserSWIVector;
    end;

    WriteMCUReg(LocalFrame);

{Loop until the MCU signals that a breakpoint has been reached}
    while (not Application.Terminated) and (Prog.State = psTrace) do
      Application.ProcessMessages;

{Update StartRecord and loop back}
    StartRecord := BreakRecord;
    with StartRecord do
    begin
      BPType         := bpStart;
      IsNextTraceBP  := false;
      ActingAsTempBP := false;
    end;

  end; // for i := 1 to Repeats do
end; // procedure Trace

procedure Go;
var
  LocalFrame       : tStackFrame;
  BPptr            : tBPRecordPtr;
  i                : integer;
  StoppedAtBP,Flag : boolean;
begin
  SwapTVectorsIn;
  if not InitialReadRegs
  then begin
    ReadMCUReg(LocalFrame);
    Prog.StackFrame := LocalFrame;
  end;

  RemoveTransientSetByTrace;
  RemoveTraceRequests;
  DeleteBPbyType([bpStart]);
  StoppedAtBP := StartAddr = Prog.BRAddr;
  Prog.BRAddr := $FFFF;

{Add any new breakpoints}
  for i := 0 to length(ATArray) - 1 do
  begin
    MakeBPRecord(ATArray[i].Addr,ATArray[i].BPTYpe,BreakRecord);
    AddBPtoList(BreakRecord);
  end;

{If the breakpoint at the starting address is a transient one, then delete it}
  DeleteBPatAddr(StartRecord.Addr,[bpTransient]);

{If starting at a Fixed or Pass breakpoint, establish the necessary Temp
 breakpoint}
  if RetrieveBP(StartRecord.Addr,[bpFixed,bpPass],BPptr)
  then begin
    MakeAndAddTempBP(BPptr^,Prog.StackFrame);
  end;

  AddBPtoList(StartRecord);
  UpdateBPWindow;
  SetBreaksX([bpTemp..bpPass],StartRecord.Addr);
  CommsForm.ComPort.OnRxChar := CommsForm.ComPortRxChar1;
  Prog.State := psRun;
  PauseEndSignalled := false;

  MainForm.PCLbl.Caption := IntToHex(StartAddr,4);
  LocalFrame             := Prog.StackFrame;

  if ProjData.ShowOnceOnly
  then Flag := StartAddrSpecified
  else Flag := true;
  if Flag
  then begin
    if ProjData.ShowStartInst
    then MainForm.OWEdit.Lines.Add('Start Inst: ' + StartRecord.OWStr);
    if ProjData.ShowStartRegs
    then MainForm.OWEdit.Lines.Add('Start Regs: ' + FrameToString);
  end;
  StartAddrSpecified := false;

  if StartRecord.InstType = itSWI
  then begin
    EmulateSWI(StartAddr,LocalFrame,StoppedAtBP);
    LocalFrame.PC := UserSWIVector;
  end;

//  ShowMessage('Ready to Roll');

  WriteMCUReg(LocalFrame);
end; // procedure Go

procedure HandleSWI;
var
  LocalFrame   : tStackFrame;
  BPAddr       : word;
  BPPtr        : tBPRecordPtr;
  Halting      : boolean;
  UserPlacedSWI : boolean;
begin
  ServiceSWI(LocalFrame,BPAddr);
  BPPtr         := nil;
//  Halting       := true;
  Prog.BRAddr   := BPAddr;
  UserPlacedSWI := false;

{If the break address, BPAddr, is that of a JBug11-set BP, then possibly pause
 at this address... }
  if RetrieveBP(BPAddr,[bpTemp..bpPass],BPPtr)
  then begin
    BPPtr^.PassCount := BPPtr^.PassCount + 1;
    BreakRecord      := BPPtr^;

{Clear the current breakpoint whatever its kind}
    ClearBreak(BPAddr);

{Clear and delete all temporary breakpoints}
    ClearBreaksByType([bpTemp]);
    DeleteBPByType([bpTemp]);

{Is this a breakpoint at which execution should halt and control be returned
 to the user?}
    Halting := HaltQuery(BPPtr^,Prog.State);
  end // if RetrieveBP(BPAddr,[bpTemp..bpPass],BPPtr) then..

  else begin
    if UserSWIVector <> MapData.swi_srv
    then begin
      UserPlacedSWI := true;
      Halting       := false;
    end
    else begin
      MessageDlg('Unexpected SWI instruction encountered at $' +
                  IntToHex(BPAddr,4) + CRLF +
                  'Running Program will stop here.' + CRLF +
                  'See Help for more information',
                  mtError,[mbOK,mbHelp],IDH_SWI_IN_USER_CODE);
      Stop;
      exit;
    end; // if UserSWIVector <> MapData.swi_srv then.. else..
  end; // if RetrieveBP(BPAddr,[bpTemp..bpPass],BPPtr) then.. else..

{If halting, then:
  1. display breakpoints and
  2. refresh the Watch window
  3. add info to Output Window}

  if Halting
  then begin
    with Prog do
    begin
      TGAddr := BPAddr;
      LUAddr := BPAddr;
      LocalFrame.PC := BPAddr;
      StackFrame := LocalFrame;
      case State of
        psRun   : State := psRunStop;
        psTrace : State := psTraceStop;
      end; // case State of
    end;

    DiscoverNext(BPptr^,LocalFrame,false);
    UpdateBPWindow;
    RefreshWatch;

    if ProjData.ShowBreakInst
    then MainForm.OWEdit.Lines.Add('Break Inst: ' + BreakRecord.OWStr);
    if ProjData.ShowBreakRegs
    then MainForm.OWEdit.Lines.Add('Break Regs: ' + FrameToString);
  end
{If not halting:}
  else begin
    if UserPlacedSWI
    then begin
      EmulateSWI(BPAddr,LocalFrame,true);
      LocalFrame.PC := UserSWIVector;
    end
    else begin
      Prog.BRAddr := $FFFF;
      MakeAndAddTempBP(BPptr^,LocalFrame);
      SetBreaksX([bpTemp..bpPass],BPAddr);
      LocalFrame.PC := BPAddr;
    end;
//    ShowMessage('Ready to Roll');
    CommsForm.ComPort.OnRxChar := CommsForm.ComPortRxChar1;
    WriteMCUReg(LocalFrame);
  end; // if Halting then.. else..
end; // procedure HandleSWI

procedure Stop;
var
  LocalFrame  : tStackFrame;
  RunningPC   : word;
  Buf         : string;
begin
  if (not InitialReadRegs) or (Prog.State in [psRun,psTrace])
  then ReadMCUReg(LocalFrame)
  else LocalFrame := Prog.StackFrame;

  RemoveTransientSetByTrace;

  RunningPC      := LocalFrame.PC;
  LocalFrame.PC  := MapData.talker_idle;    // Set PC to talker idle address
  LocalFrame.CCR := LocalFrame.CCR and $EF; // Clear the I bit (seems to be necessary, 2006-04-12)
  WriteMCUReg(LocalFrame);                  // write to MCU

  Prog.StackFrame := LocalFrame;            // and write to the Prog copy
  Prog.BRAddr     := $FFFF;
  ClearBreaksByType([bpTemp..bpPass]);
  RemoveTraceRequests;
  DeleteBPByType([bpStart,bpTemp]);

  LookupSymbolName(RunningPC,Buf);
  MainForm.OWEdit.Lines.Add('Stopped at: ' + IntToHex(RunningPC,4) + ' ' + Buf);

  if Prog.State in [psRun,psTrace] then Prog.TGAddr := RunningPC;

  CommsForm.ComPort.OnRxChar := CommsForm.ComPortRxCharTest;
  Prog.State := psStopped;
  UpdateBPWindow;
  SwapTVectorsOut;
  VectorsRead := false;
  BprotOK     := false;
  RefreshWatch;
end; // procedure Stop

end.

