unit UnAsmUnit;

interface

uses Classes, SysUtils, TypeDefs;

procedure DisAsm1(    DAddr      : word;
                  var BPRecord   : tBPRecord);
function  UnassembleCheck(const CList   : tStringList;
                          var   Comment : tCmdLineComment;
                          var   Command : tCommand) : boolean;
procedure Unassemble;
procedure UnassembleLocal;

implementation

uses Main, CmdUnit, FileUnit, LookupUnit, MemRW, CheckUnit, MiscUnit, InitUnit;

var
  StartAddr,EndAddr : word;
  Repeats           : integer;

{DisAsm1  ******************************************************************
 Routine to dis-assemble a single instruction.  The necessary data bytes
 must already be present in the Memory array, with DAddr pointing to the first
 byte of the instruction.  For an unknown instruction (the usual case) this
 implies that at least the four bytes following the one at DAddr in the local
 memory array must be a fair copy of the MCU memory.

 BPRecord is filled in with all the necessary data for a breakpoint
 when the function returns.

 SeqAddr is the address of the next instruction in sequence in memory, not
 necessarily the next to be executed.  If the instruction is a conditional
 branch, then AltAddr holds the address for the case that the condition is
 satisfied.

 InstType holds an indicator of the type of instruction.  In the case
 of branch, indirectly addressed jumps, and RTI and RTS instructions,
 external processing is required to determine the next instruction
 address - see comments in 'DiscoverNext' in the RunTraceUnit}

procedure DisAsm1(    DAddr      : word;
                  var BPRecord   : tBPRecord);
var
  i                              : integer;
  Buf,BranchAddrStr,LabelAddrStr : string;
  AddrH,AddrL,w                  : word;
  TPtr                           : tOpcodeDatumPtr;
  LongAddr                       : longword;
begin
  Buf := '';
  fillchar(BPRecord,sizeof(BPRecord),0);
  with BPRecord,ProjData do
  begin
{SeqAddr and AltAddr are set to the impossible value of $FFFF as a safeguard,
 and as a flag to the trace routine that no breakpoint can be set here}
    AddrL    := $FFFF;
    AddrH    := $FFFF;
    Addr     := DAddr;
    SeqAddr  := $FFFF;
    AltAddr  := $FFFF;
    NextAddr := $FFFF;
    MemVal   := Memory[DAddr];    // used by Set and Clear breakpoints
    InstType := itNone;
    Offset   := 0;

{Print the disassembly address to the OW and Breakpoint Window display strings}
    OWStr       := format('%4.4x',[DAddr]) + '  ';
    BPStr       := format('%-5.4x',[DAddr]);

{Add the first byte of the instruction to the output strings}
    OWStr := OWStr + format('%2.2x',[Memory[DAddr]]);
    BPStr := BPStr + format('%-3.2x',[Memory[DAddr]]);

{Fetch the pointer to the opcode de-coding information}
    case Memory[DAddr] of
      $18 : TPtr := OpcodePage2[Memory[DAddr+1]];
      $1A : TPtr := OpcodePage3[Memory[DAddr+1]];
      $CD : TPtr := OpcodePage4[Memory[DAddr+1]]
    else TPtr := OpcodePage1[Memory[DAddr]];
    end;

{if the decoding info is not found then opcode must be illegal or a data byte}
    if TPtr = nil
    then begin
      OWStr    := OWStr + ^I^I + '; data or illegal opcode';
      BPStr    := BPStr + ';illegal opcode';
      InstType := itIllegalOpcode;
      SeqAddr  := DAddr + 1;
      exit;
    end;

{Copy Tptr to the BPRecord}
    OpcodeDatumPtr := TPtr;

    with TPtr^ do
    begin
{Generate sequential address}
      LongAddr := DAddr + PCOffset;       // LongAddr is a temorary variable
      if LongAddr > $FFFF then exit else SeqAddr := LongAddr;

{Add the bytes which make up this instruction to the OW string}
      for i := 1 to PCOffset - 1 do
      begin
        OWStr := OWStr + format('%2.2x',[Memory[DAddr + i]]);
      end;

{Adjust OW string length}
      SpaceTo(OWStr,17);

{if the address corresponds to a symbol, include the symbol}
      if (SymbolTableToUse <> stNone) and LookupSymbolName(DAddr,LabelAddrStr)
      then OWStr := OWStr + LabelAddrStr;

{Adjust OW string length}
      SpaceTo(OWStr,33);

{Advance DAddr if a page 2, 3 or 4 opcode}
      if Memory[DAddr] in [$18,$1A,$CD] then DAddr   := DAddr + 1;

{Read in the three (max) operands}
      Opd1 := Memory[DAddr + 1];
      Opd2 := Memory[DAddr + 2];
      Opd3 := Memory[DAddr + 3];

{Add Mnemonic to display strings}
//      OWStr := OWStr + ^I;                        // Add one or two tabs
//      if PCOffset < 4 then OWStr := OWStr + ^I;   
      OWStr := OWStr + Mnemonic;
      SpaceTo(OWStr,41);
      BPStr := BPStr + Mnemonic + ' ';

{Build the display strings}
      case Opd1Type of
        otNone: begin end;
        otII  : Buf := Buf + '#$' + IntToHex(Opd1,2);
        otJJ  : begin
                  AddrH := Opd1;
                  Buf := Buf + '#';
                end;
        otHH  : AddrH := Opd1;
        otDD  : begin
                  AddrH  := 0;
                  AddrL  := Opd1;
                end;
        otRR  : begin
                  AltAddr := word(smallint(SeqAddr) + shortint(Opd1));
                  if (SymbolTableToUse <> stNone) and
                      LookupLabelName(AltAddr,BranchAddrStr)
                  then Buf := Buf + BranchAddrStr
                  else Buf := Buf + '$' + IntToHex(AltAddr,4);
                end;
        otFF  : begin
                  Buf := Buf + '$' + IntToHex(Opd1,2);
                  case AddrMode of
                    amIX : Buf := Buf + ',X';
                    amIY : Buf := Buf + ',Y';
                  end;
                end;
      end; // case Opd1Type of

      if Opd2Type in [otLL,otKK] then AddrL := Opd2;

{If AddrH and AddrL have both been modified, then see if they are a label}
      if (AddrH < $100) and (AddrL < $100)
      then begin
        w := (AddrH shl 8) + AddrL;
        if (SymbolTableToUse <> stNone) and
           LookupLabelName(w,LabelAddrStr)
        then Buf := Buf + LabelAddrStr
        else Buf := Buf + '$' + IntToHex(w,4);
      end;

      if Opd2Type = otMM then Buf := Buf + ',%' + WordToBinStr(Opd2,8);

      if Opd3Type = otRR
      then begin
        AltAddr := word(smallint(SeqAddr) + shortint(Opd3));
        if (SymbolTableToUse <> stNone) and
            LookupLabelName(AltAddr,BranchAddrStr)
        then Buf := Buf + ',' + BranchAddrStr
        else Buf := Buf + ',' + '$' + IntToHex(AltAddr,4);
      end;

      OWStr := OWStr + Buf;
      BPStr := BPStr + Buf;

{Determine instruction type *************************************************}
{tInstructionType = (itNone,itIllegalOpcode,itSimple,itBranch,itBSR,itJSR,
                     itJump,itJumpIX,itJumpIY,itRTI,itRTS,itSWI);          }
{tAddrMode = (amNone,amINH,amIMM,amDIR,amEXT,amIX,amIY,amREL,amDirective); }

      InstType := itSimple;      // default

      if (Opd1Type = otRR) or (Opd3Type = otRR)
      then InstType := itBranch;

      if Mnemonic[1] = 'J'       // if the instruction is an unconditional jump
      then begin
        case AddrMode of
          amDIR,amEXT : begin
                          AltAddr  := (AddrH shl 8) + AddrL;
                          InstType := itJump;
                        end;
          amIX        : InstType := itJumpIX;
          amIY        : InstType := itJumpIY;
        end;
      end;

      if Mnemonic = 'BSR' then InstType := itBSR;
      if Mnemonic = 'JSR' then InstType := itJSR;
      if Mnemonic = 'RTI' then InstType := itRTI;
      if Mnemonic = 'RTS' then InstType := itRTS;
      if Mnemonic = 'SWI' then InstType := itSWI;

      Inst := Mnemonic;
    end; // with TPtr^ do
  end; // with BPRecord do
end; // function DisAsm1

function  UnassembleCheck(const CList   : tStringList;
                          var   Comment : tCmdLineComment;
                          var   Command : tCommand) : boolean;
var
  A1,A2,Val : longword;
  ArgType  : tArgType;
  Buf      : string;
  IsLocal  : boolean;
begin
  Result  := false;
  Comment := clcArgErr;
  A1 := 0; A2 := 0;
  IsLocal := uppercase(copy(CList[0],2,1)) = 'L';

  if not NumArgsOK(CList,0,2,Comment) then exit;

  if not IsLocal and not
     StateIsAllowed([psStopped,psRun,psRunStop,psTraceStop],Comment)
  then exit;

  Buf := uppercase(CList[0]);
  if (Buf = 'U') or (Buf = 'UL')
  then Repeats := ProjData.LinesToUnassemble
  else begin
    if copy(Buf,2,1) = 'L' then delete(Buf,1,2) else delete(Buf,1,1);
    try
      Repeats := StrToInt(Buf);
    except on EConvertError do
      begin
        Comment := clcCmdErr;
        exit;
      end;
    end; // try.. except on EConvertError do
  end; // if (Buf = 'U') or (Buf = 'UL') then.. else..

  if FoundArgOK(CList,1,[atNotExist,atByte,atWord,atLabel,atAsterisk],
                ArgType,Val)
  then begin
    case ArgType of
      atNotExist            : A1 := Prog.LUAddr;
      atAsterisk            : A1 := Prog.StackFrame.PC;
      atByte,atWord,atLabel : A1 := Val;
    end;
  end else exit;

  if FoundArgOK(CList,2,[atNotExist,atByte,atWord,atLabel,atLength],
                ArgType,Val)
  then begin
    case ArgType of
      atNotExist            : A2 := A1;
      atByte,atWord,atLabel : A2 := Val;
      atLength              : A2 := A1 + Val;
    end;
  end else exit;

{'Repeats' may not be specified as more than 1 and at the same time specify
 a second address}
    if ArgType <> atNotExist then Repeats := 1;

{Check that: 1. A2 is larger than A1
             2. A2 <= $FFFF
             3. (No address range exclusions apply to the list command) }
  if (A2 < A1) or (A2 > $FFFC)
  then begin
    Comment := clcAddrRangeErr;     // Address range error
    exit;
  end;
  StartAddr := A1;
  EndAddr   := A2;
  Comment   := clcOK;
  Result    := true;  // Arguments are OK
  if IsLocal then Command := cmdUnassembleLocal else Command := cmdUnassemble;
end; // function UnassembleCheck

procedure UnassembleLocal;
var
  i     : integer;
//---------------------------------
  procedure ProcessLine;
  var BPRec : tBPRecord;
  begin
    DisAsm1(StartAddr,BPRec);
    StartAddr := BPRec.SeqAddr;
    with BPRec,MainForm.OWEdit do
    begin
      Lines.Add(format('%-4.4d',[i]) + ':  ' + OWStr);
      if ProjData.AddBreakSpaces
      then begin
        if (  InstType in [itJumpIX,itJumpIY,itRTI,itRTS,itSWI] ) or
           ( (InstType in [itJump]) and (AltAddr <> SeqAddr)    ) or
           ( (Inst = 'BRA') and (AltAddr <> SeqAddr)            )
        then Lines.Add('');
      end;
    end;
  end; // procedure ProcessLine
//---------------------------------
begin
  if Repeats > 1
  then begin
    for i := 1 to Repeats do ProcessLine;
  end
  else begin
    i := 1;
    while StartAddr <= EndAddr do
    begin
      ProcessLine;
      i := i + 1;
    end;
  end; // if Repeats > 1 then.. else..
  Prog.LUAddr := StartAddr;
end; // procedure UnassembleLocal

procedure Unassemble;
var
  LastAddr : longword;
begin
  if EndAddr = StartAddr
  then begin
    if Repeats = 1
    then ReadMCU64K(StartAddr,StartAddr + 4,false)
    else begin
      LastAddr := StartAddr + 2*Repeats + 10;
      if LastAddr > $FFFF then LastAddr := $FFFF;
      ReadMCU64K(StartAddr,LastAddr,true);
    end; // if Repeats = 1 then.. else..
  end
  else begin
    ReadMCU64K(StartAddr,EndAddr,true);
  end; // if EndAddr = StartAddr then.. else..
  UnassembleLocal;
end; // procedure Unassemble

end.


