unit VIUnit;
{ Defines a class TVersionInfo which returns the Project Version Info strings,
  the Locale ID converted to a Language ID string, plus the Major, Minor,
  Release versions and Build number

  By David Simpson (drs@ihug.com.au)
  3/8/99

  This unit is based on code originally supplied by Josef Garvi
  (josef@eden-foundation.org).  The key difference between his implementation
  and mine is that his returned values as procedural parameters while mine
  does so as properties. }

interface

uses Classes, Forms, Windows, Dialogs, SysUtils;

type
  TVersionInfo = class(TObject)
  private
    FileName : string;
    // for GetVersionString
    FVData : Pointer;
    FVSize : LongInt;
    FVVers : Pointer;
    FLen : DWord;
    FLangIDStr : string;
    FPLDescr : PChar;
    // for GetVersionNumber
    FVerInfo : Pointer;
    FVerInfoSize : DWord;
    FVerValue : PVSFixedFileInfo;
    function GetVersionString(Index : integer) : string;
    function GetVersionNumber(Index : integer) : word;
  public
    constructor Create(Owner : TObject);
    destructor Destroy; override;
    property Language : string index 0 read GetVersionString;
    property CompanyName : string index 1 read GetVersionString;
    property FileDescription : string index 2 read GetVersionString;
    property FileVersion : string index 3 read GetVersionString;
    property InternalName : string index 4 read GetVersionString;
    property LegalCopyright : string index 5 read GetVersionString;
    property LegalTrademarks : string index 6 read GetVersionString;
    property OriginalFileName : string index 7 read GetVersionString;
    property ProductName : string index 8 read GetVersionString;
    property ProductVersion : string index 9 read GetVersionString;
    property Comments : string index 10 read GetVersionString;
    property MajorVersion : word index 0 read GetVersionNumber;
    property MinorVersion : word index 1 read GetVersionNumber;
    property ReleaseVersion : word index 2 read GetVersionNumber;
    property Build : word index 3 read GetVersionNumber;
  end;

var
  VersionInfo : tVersionInfo;

implementation

constructor TVersionInfo.Create(Owner : TOBject);
var
  // for GetVerString
  VIHandle : DWord;
  LangID : DWord;
  LDescr : array[0..255] of Char;
  // for GetVersionNumber
  VerValueSize : DWord;
  Dummy : DWord;

  function GetLanguageID() : DWord;
    { Retrieves the language ID, needed to know the locale / language code needed
      to retrieve the specific strings. }
  const
    Address = '\VarFileInfo\Translation';
  var
    s : string;
    wArr : array of Word;
  begin
    Result := 0;
    s := Address;
    if VerQueryValue(FVData, PChar(s), FVVers, FLen) then
      if FLen > 1 then
        begin
          FLen := FLen div SizeOf(Word) * SizeOf(Word);
          SetLength(wArr, FLen div SizeOf(Word));
          Move(FVVers^, wArr[0], FLen);
          Result := wArr[1];
          Result := Result or (wArr[0] shl 16)
        end;
  end;

begin
  inherited Create;
  FileName := Application.ExeName;
  // for GetVersionString
  FVSize := GetFileVersionInfoSize(PChar(FileName), VIHandle);
  if FVSize <> 0 then
    begin
      GetMem(FVData, FVSize);
      if GetFileVersionInfo(@FileName[1], VIHandle, FVSize, FVData) then
        begin
          LangID := GetLanguageID;
          FLangIDStr := IntToHex(LangID, 8);
          FPLDescr := LDescr;
          VerLanguageName(HiWord(LangID), FPLDescr, 256);
        end
    end;
  // for GetVersionNumber
  FVerInfoSize := GetFileVersionInfoSize(PChar(FileName), Dummy);
  if FVerInfoSize = 0 then
    begin
      Dummy := GetLastError;
      ShowMessage('GetFileVersionInfoSize error #:' + IntToStr(Dummy));
    end;
  GetMem(FVerInfo, FVerInfoSize);
  GetFileVersionInfo(PChar(FileName), 0, FVerInfoSize, FVerInfo);
  VerQueryValue(FVerInfo, '\', Pointer(FVerValue), VerValueSize);
end;

destructor TVersionInfo.Destroy;
begin
  // for GetVersionString
  FreeMem(FVData, FVSize);
  // for GetVersionNumber
  FreeMem(FVerInfo, FVerInfoSize);
  inherited Destroy;
end;

function TVersionInfo.GetVersionNumber(Index : integer) : word;
{ Returns the Major, Minor, Release versions and Build number }

begin
  // dummy code to stop compiler warning
  Result := 0;
  // real code
  case Index of
    0 : Result := FVerValue^.dwFileVersionMS shr 16; // Major Version
    1 : Result := FVerValue^.dwFileVersionMS and $FFFF; // Minor Version
    2 : Result := FVerValue^.dwFileVersionLS shr 16; // Release Version
    3 : Result := FVerValue^.dwFileVersionLS and $FFFF; // Build
  end;
end;

function TVersionInfo.GetVersionString(Index : integer) : string;
{ Returns the Project Version Info strings plus the Locale ID converted to
  a Language ID string. }

  function GetFVerValue(Value : string) : string;
    // Retrieves specific strings
  var
    ItemName : string;
  const
    Prefix = '\StringFileInfo\';
  begin
    ItemName := Prefix + FLangIDStr + '\' + Value;
    Result := '';
    if VerQueryValue(FVData, PChar(ItemName), FVVers, FLen) then
      if FLen > 1 then
        begin
          if FLen > 255 then FLen := 255; { "Truncate" any long strings }
          SetLength(Result, FLen - 1);
          Move(FVVers^, Result[1], FLen - 1);
        end;
  end;

begin
  case Index of
    0 : Result := FPLDescr; // Language
    1 : Result := GetFVerValue('CompanyName');
    2 : Result := GetFVerValue('FileDescription');
    3 : Result := GetFVerValue('FileVersion');
    4 : Result := GetFVerValue('InternalName');
    5 : Result := GetFVerValue('LegalCopyright');
    6 : Result := GetFVerValue('LegalTrademarks');
    7 : Result := GetFVerValue('OriginalFilename');
    8 : Result := GetFVerValue('ProductName');
    9 : Result := GetFVerValue('ProductVersion');
    10 : Result := GetFVerValue('Comments');
  end;
end;

end.


